class baseChains(metaclass=abc.ABCMeta):
    def __init__(self, targetDist):
        self.stepNum = 0
        self.targetDist = targetDist
        self.tryplot()

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "reset")
            and callable(subclass.reset)
            and hasattr(subclass, "step")
            and callable(subclass.step)
            or NotImplemented
        )

    @abc.abstractmethod
    def reset(self, x):
        """Return to initial MVN"""
        raise NotImplementedError

    @abc.abstractmethod
    def step(self, x):
        """Take a sampling step"""
        raise NotImplementedError

    def tryplot(self):
        """Determine if there is any point setting up plotting environments"""
        import sys
        import importlib

        if (
            "matplotlib" in sys.modules
            or (spec := importlib.util.find_spec("matplotlib")) is not None
        ):
            self.plotPrepareDefaults()
        return

    def plotPrepareDefaults(self):
        """Better defaults than normal"""
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        mpl.rcParams["figure.figsize"] = [12, 12]
        mpl.rcParams["figure.dpi"] = 72

        plt.style.use("fivethirtyeight")
        plt.rcParams["font.size"] = 14
        plt.rcParams["axes.labelsize"] = 14
        plt.rcParams["axes.labelweight"] = "bold"
        plt.rcParams["axes.titlesize"] = 14
        plt.rcParams["xtick.labelsize"] = 10
        plt.rcParams["ytick.labelsize"] = 10
        plt.rcParams["legend.fontsize"] = 14
        plt.rcParams["figure.titlesize"] = 16

        width, height = plt.figaspect(1.68)
        fig = plt.figure(figsize=(width, height), dpi=400)

    def computeMean(self, chain):
        """Computes the chain mean"""
        return np.mean(chain)  # TODO: figure this out

    def computeAutocorrelation(self, lagRange=201):
        """Get the autocorrelation"""

        def acf(x, lagRange):
            return np.array(
                [1]
                + [np.corrcoef(x[:-i], x[i:])[0, 1] for i in range(1, lagRange)]
            )

        dat = self.extractAllXY()
        return (acf(dat.x, lagRange), acf(dat.y, lagRange))

    @functools.cached_property
    def extractDensity(
        self, xlim={"low": -8, "high": 10}, ylim={"low": -15, "high": 4}
    ):
        return self.targetDist.plotDensity

    def extractXYSamples(self, accepted=True, burnin=500):
        """Extract a pandas dataframe of the x,y pairs"""
        assert self.stepNum > burnin, f"Step must be greater than {burnin}"
        assert self.dim == 2, f"Samples only for dim == 2, got {self.dim}"
        if accepted == True:
            dat = pd.DataFrame(
                [
                    (x.proposal[0], x.proposal[1])
                    for x in self.traj[burnin:-1]
                    if x.acceptance
                ]
            )
        else:
            dat = pd.DataFrame(
                [
                    (x.proposal[0], x.proposal[1])
                    for x in self.traj[burnin:-1]
                    if not x.acceptance
                ]
            )
        dat.columns = ["x", "y"]
        return dat

    def extractAllXY(self, burnin=500):
        dat = pd.DataFrame(
            [(x.proposal[0], x.proposal[1]) for x in self.traj[burnin:-1]]
        )
        dat.columns = ["x", "y"]
        return dat

    def plotTargetSamples(
        self,
        xlim={"low": -8, "high": 10},
        ylim={"low": -15, "high": 4},
        title="",
    ):
        """Overlay samples on a target distribution"""
        paccepted = self.extractXYSamples(accepted=True)
        prejected = self.extractXYSamples(accepted=False)
        res = self.extractDensity(xlim=xlim, ylim=ylim)
        fig = plt.figure()
        plt.contour(res.xx, res.yy, res.zz)
        plt.plot(prejected.x, prejected.y, "ro", alpha=0.4, label=r"rejected")
        plt.plot(paccepted.x, paccepted.y, "bo", alpha=0.5, label=r"accepted")
        plt.legend()
        plt.title(title + "- Samples")
        plt.xlabel("X")
        plt.ylabel("Y")
        return fig

    def plotTrace(self, burnin=500, title=""):
        """Generate a trace plot"""
        dat = self.extractAllXY(burnin=burnin)
        fig = plt.figure()
        plt.plot(np.arange(len(dat.x)), dat.x, label=r"$x$")
        plt.plot(np.arange(len(dat.y)), dat.y, label=r"$y$")
        plt.legend()
        plt.title(title + "- Trace Plot")
        plt.xlabel("Step")
        plt.ylabel("Trace")
        plt.xlim([0, self.stepNum - burnin])
        return fig

    def plotAutocorrelations(self, lagRange=301, title=""):
        acx, acy = self.computeAutocorrelation(lagRange)
        fig = plt.figure()
        kplot = np.arange(lagRange)
        plt.plot(kplot[1:], acx[1:], label=r"$x$")
        plt.plot(kplot[1:], acy[1:], label=r"$y$")
        plt.legend()
        plt.title(title + "- Samples")
        plt.xlabel("X")
        plt.ylabel("Y")
        return fig


class RandomWalkMetropolisHastingsMC(baseChains):
    def __init__(
        self,
        dimensionsLike,
        targetDist,
        sigma=1,
    ):
        super().__init__(targetDist=targetDist)
        self.sigma = sigma
        self.dim = dimensionsLike.shape[0]  # np.zeros(dim)
        self.chain = [pbsd.mvn.MultivariateNormal.getMVNSample(self.dim)]
        self.traj = []

    def reset(self):
        self.chain = [pbsd.mvn.MultivariateNormal.getMVNSample(self.dim)]
        self.stepNum = 0
        self.traj = []

    def step(self):
        isaccept = None
        proposalDist = pbsd.mvn.MultivariateNormal(
            self.chain[-1], np.eye(self.dim) * self.sigma**2
        )
        proposal = proposalDist.getSample()
        logAcceptRatio = self.targetDist.logDensity(
            proposal
        ) - self.targetDist.logDensity(self.chain[-1])
        if np.random.default_rng().uniform() < np.exp(logAcceptRatio):
            self.chain.append(proposal)
            isaccept = True
        else:
            self.chain.append(self.chain[-1])
            isaccept = False
        self.traj.append(
            aux.structs.mcmcData(
                step=self.stepNum,
                acceptance=isaccept,
                proposal=copy.deepcopy(proposal),
                proposalDistCovariance=proposalDist.covMat,
            )
        )
        self.stepNum += 1


class HamiltonianMC(baseChains):
    def __init__(self, dimensionsLike, targetDist, leapfrogSteps=37, dt=0.1):
        super().__init__(targetDist=targetDist)
        self.leapfrogSteps = 37
        self.dt = 0.1
        self.dim = dimensionsLike.shape[0]  # np.zeros(dim)
        self.targetDist = targetDist
        self.stepNum = 0
        self.petmvn = pbsd.mvn.MultivariateNormal(np.zeros(2), np.eye(2))
        self.chain = [self.petmvn.getSample()]
        self.traj = []

    def reset(self):
        self.petmvn = pbsd.mvn.MultivariateNormal(np.zeros(2), np.eye(2))
        self.chain = [self.petmvn.getSample()]
        self.stepNum = 0
        self.traj = []

    def step(self):
        isaccept = None
        currentPositions = self.chain[-1]
        currentMomenta = pbsd.mvn.MultivariateNormal.getMVNSample(self.dim)
        if self.stepNum == 0:
            self.traj.append(
                aux.structs.mcmcData(
                    step=self.stepNum,
                    acceptance=True,
                    proposal=currentPositions,
                    proposalDistCovariance=None,
                )
            )
            self.stepNum += 1
            return
        # Otherwise..
        propMomenta = currentMomenta
        propPositions = currentPositions
        # Leapfrog
        # Half step
        propMomenta -= (
            self.targetDist.gradLogDensity(propPositions) * self.dt / 2.0
        )
        # Full steps for position and momentum
        for istep in range(self.leapfrogSteps):
            propPositions += propMomenta * self.dt
            if istep != self.leapfrogSteps - 1:
                propMomenta -= (
                    self.targetDist.gradLogDensity(propPositions) * self.dt
                )
        # Half step
        propMomenta -= (
            self.targetDist.gradLogDensity(propPositions) * self.dt / 2.0
        )

        # Acceptance ratio
        currentHamiltonian = self.targetDist.logDensity(
            currentPositions
        ) + self.targetDist.logDensity(currentMomenta)
        propHamiltonian = self.targetDist.logDensity(
            propPositions
        ) + self.targetDist.logDensity(propMomenta)
        logAcceptRatio = propHamiltonian - currentHamiltonian
        # Determine outcome
        if np.random.default_rng().standard_normal() < np.exp(logAcceptRatio):
            self.chain.append(propPositions)
            isaccept = True
        else:
            self.chain.append(currentPositions)
            isaccept = False
        self.traj.append(
            aux.structs.mcmcData(
                step=self.stepNum,
                acceptance=isaccept,
                proposal=copy.deepcopy(propPositions),
                proposalDistCovariance=None,
            )
        )
        self.stepNum += 1



from collections import namedtuple

drawvals = namedtuple("drawvals", ["res", "x", "y"])
plotvals = namedtuple("plotvals", ["xx", "yy", "zz"])
mcmcData = namedtuple(
    "mcmcData", ["step", "acceptance", "proposal", "proposalDistCovariance"]
)
mcmcSamples = namedtuple("mcmcSamples", ["acceptedVals", "rejectedVals"])



# Common
import abc
import numpy as np
import scipy
from collections import namedtuple

# Internal libraries
import probsamplers.distributions as pbsd
from probsamplers import aux


class baseTargetDistrib(metaclass=abc.ABCMeta):
    def __init__(self, xmin, xmax):
        self.xmin = xmin
        self.xmax = xmax

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "logDensity")
            and callable(subclass.logDensity)
            and hasattr(subclass, "gradLogDensity")
            and callable(subclass.gradLogDensity)
            or NotImplemented
        )

    @abc.abstractmethod
    def logDensity(self, x):
        """Return the log of the probability density"""
        raise NotImplementedError

    @abc.abstractmethod
    def gradLogDensity(self, x):
        """Returns the gradient of the probability density"""
        raise NotImplementedError

    @abc.abstractmethod
    def gradLogDensity(self, x):
        """Returns the gradient of the probability density"""
        raise NotImplementedError

    def plotDensity(
        self,
        xlim={"low": -1.5, "high": 1.5},
        ylim={"low": -1.5, "high": 1.5},
        nstep=200,
    ):
        x = np.linspace(xlim["low"], xlim["high"], nstep)
        y = np.linspace(ylim["low"], ylim["high"], nstep)
        xx, yy = np.meshgrid(x, y)
        zz = np.zeros((nstep, nstep))
        for i in np.arange(0, nstep):
            for j in np.arange(0, nstep):
                zz[i, j] = np.exp(self.logDensity([xx[i, j], yy[i, j]]))
        return aux.structs.plotvals(xx=xx, yy=yy, zz=zz)


class rosenbrockBanana(baseTargetDistrib):
    def __init__(
        self,
        xmin=-6,
        xmax=6,
        alpha=2,
        beta=0.2,
        mu=np.array([0, 4]),
        cov=np.array([1, 0.5, 0.5, 1]).reshape(2, 2),
    ):
        super().__init__(xmin, xmax)
        self.alpha = alpha
        self.beta = beta
        self.pdistrib = pbsd.mvn.MultivariateNormal(mu, cov)

    def getYvec(self, x):
        assert len(x) == 2
        yvec = np.zeros(2)
        yvec[0] = x[0] / self.alpha
        yvec[1] = x[1] * self.alpha + self.alpha * self.beta * (
            x[0] ** 2 + self.alpha**2
        )
        return yvec

    def logDensity(self, x):
        """Return the log probability of the Banana function"""
        return self.pdistrib.logDensity(self.getYvec(x))

    def gradLogDensity(self, x):
        y = self.getYvec(x)  # Uses x[1]
        grad = self.pdistrib.gradLogDensity(y)
        gradx0 = (
            grad[0] / self.alpha + grad[1] * self.alpha * self.beta * 2 * x[0]
        )
        gradx1 = grad[1] * self.alpha
        return np.array([gradx0, gradx1])


class donutDistrib(baseTargetDistrib):
    def __init__(self, xmin=-6, xmax=6, radius=2.6, sigma2=0.033):
        super().__init__(xmin, xmax)
        self.radius = radius
        self.sigma2 = sigma2

    def logDensity(self, x):
        rval = np.linalg.norm(x)
        return -1 * np.power(rval - self.radius, 2) / self.sigma2

    def gradLogDensity(self, x):
        rval = np.linalg.norm(x)
        return np.array([])
        gradx0 = x[0] * (self.radius / (rval - 1)) / self.sigma2
        gradx1 = x[1] * (self.radius / (rval - 1)) / self.sigma2
        return np.array([gradx0, gradx1])


class funnelDistrib(baseTargetDistrib):
    def __init__(self, xmin=-6, xmax=6):
        super().__init__(xmin, xmax)

    def f(self, x, m, s):
        return (
            -0.5 * np.log(2.0 * np.pi)
            - np.log(s)
            - 0.5 * np.power((np.array(x) - m) / s, 2)
        )

    def dfdx(self, x, m, s):
        return -(x - m) / np.power(s, 2)

    def dfds(self, x, m, s):
        return np.power(x - m, 2) - np.power(s, 2) / np.power(s, 3)

    def logDensity(self, x_):
        x = np.array([x_[1] - 2, x_[0]])
        m0 = m1 = 0
        s0 = 3
        s1 = np.exp(x[0] / 2)
        return self.f(x[0], m0, s0) + self.f([1], m1, s1)

    def gradLogDensity(self, x_):
        x = np.array(x_[1] - 2, x_[0])
        m0 = m1 = 0
        s0 = 3
        s1 = np.exp(x[0] / 2)
        return np.array(
            [
                self.dfdx(x[1], m1, np.exp(x[0] / 2)),
                self.dfds(x[0], m0, s0) + np.exp(x[0], 2),
            ]
            * self.dfds(x[1], m1, np.exp(x[0] / 2))
        )



from setuptools import setup, find_packages

setup(
    name='probsamplers',
    version='0.1.0',
    description='JupyterLab version of mcmc-demo',
    author='Rohit Goswami',
    author_email='rog32@hi.is',
    url='https://rgoswami.me',
    packages=find_packages(include=['probsamplers', 'probsamplers.*']),
    install_requires=[
        'numpy',
        'scipy'
    ],
    extras_require={'plotting': ['matplotlib>=2.2.0', 'jupyter']},
    setup_requires=['pytest-runner', 'flake8'],
    tests_require=['pytest'],
)



import numpy as np
from probsamplers import aux
import scipy.stats


class MultivariateNormal:
    # Static member
    rng = np.random.default_rng()

    def __init__(self, mu, covMat):
        self.mean = np.asarray(mu)
        self.dim = self.mean.shape[0]
        self.constant = -0.5 * np.log(2.0 * np.pi) * self.dim
        assert covMat.shape == (self.dim, self.dim)
        self.covMat = covMat
        self.covL = np.linalg.cholesky(covMat)
        self.logDet = np.log(self.covL.diagonal()).sum()

    @staticmethod
    def getNormal():
        def draw():
            pv = lambda: MultivariateNormal.rng.standard_normal() * 2 - 1
            x = pv()
            y = pv()
            return aux.structs.drawvals(res=(x**2 + y**2), x=x, y=y)

        w = draw()
        while w.res >= 1.0:
            w = draw()
        return w.x * np.sqrt((-2 * np.log(w.res)) / w.res)

    @staticmethod
    def getMVNSample(dim):
        dist = MultivariateNormal(np.zeros(dim), np.eye(dim))
        return dist.getSample()

    def getSample(self):
        return MultivariateNormal.rng.multivariate_normal(
            self.mean, self.covMat
        )

    def logDensity(self, x):
        diff = self.mean - x
        eqsolv = np.linalg.solve(self.covL, diff)
        return self.constant - np.linalg.norm(
            self.logDet - 0.5 * np.linalg.solve(self.covL, eqsolv)
        )

    def gradLogDensity(self, x):
        diff = self.mean - x
        eqsolv2 = np.linalg.solve(self.covL, diff)
        return -1 * np.linalg.solve(self.covL, eqsolv2)

    def __repr__(self):
        return f"mu: {self.mean} \t logDet: {self.logDet} \n cov: {self.covMat}\n covL: {self.covL}\n"
